"""
Date:   08/27/2023
Author: Martin E. Liza
File:   aerodynamics.py
Def:    Contains aerodynamics helper functions.
"""

import molmass
import numpy as np
import scipy.constants as s_consts
from haot import constants as constants_tables


def sutherland_law_viscosity(temperature_K: float, molecule: str = "Air") -> float:
    """
    Calculates the Sutherland's law of viscosity

    Parameters:
        temperature_K: reference temperature in [K]
        molecule: Air (default), Argon, N2, O2

    Returns:
        dynamic viscosity in [kg/ms]

    Examples:
        >> sutherland_law_viscosity(300.0)

    Reference:
        Viscous Fluid Flow, International Edition, 4th (White F., ISBN 978 1 260 59786)
    """
    const = constants_tables.sutherland_constants(molecule)

    # Eq 1-34
    dynamic_viscosity = const["temperature_ref"] + const["sutherland_visc"]
    dynamic_viscosity /= temperature_K + const["sutherland_visc"]
    dynamic_viscosity *= (temperature_K / const["temperature_ref"]) ** (3 / 2)

    return const["viscosity_ref"] * dynamic_viscosity  # [kg/ms]


def sutherland_law_conductivity(temperature_K: float, molecule: str = "Air") -> float:
    """
    Calculates the Sutherland's law of thermal conductivity

    Parameters:
        temperature_K: reference temperature in [K]
        molecule: Air (default), Argon, N2, O2

    Returns:
        thermal conductivity in [W/mK]

    Examples:
        >> sutherland_law_conductivity(300.0)

    Reference:
        Viscous Fluid Flow, International Edition, 4th (White F., ISBN 978 1 260 59786)
    """
    const = constants_tables.sutherland_constants(molecule)

    # Eq 1-41b
    thermal_conductivity = const["sutherland_cond"]
    thermal_conductivity += const["temperature_ref"]
    thermal_conductivity /= temperature_K + const["sutherland_cond"]
    thermal_conductivity *= temperature_K / const["temperature_ref"]
    thermal_conductivity **= 3 / 2

    return const["conductivity_ref"] * thermal_conductivity  # [W/mK]


def air_atomic_molar_mass(molecules: str = None) -> dict[str, float]:
    """
    Returns the atomic molar mass

    Parameters:
        molecule: Molecules that need the molar mass (11 species air is the default)

    Returns:
        species in [g/mol]

    Examples:
        >> air_atomic_molar_mass(["N+", "N2"])
    """
    if not molecules:
        molecules = ["N+", "O+", "NO+", "N2+", "O2+", "N", "O", "NO", "N2", "O2"]

    air_atomic_dict = {i: molmass.Formula(i).mass for i in molecules}

    return air_atomic_dict  # [g/mol]


def speed_of_sound(temperature_K: float, adiabatic_indx: float = 1.4) -> float:
    """
    Calculates the speed of sound

    Parameters:
        temperature_K: reference temperature in [K]
        adiabatic_indx: adiabatic index, 1.4 (default)

    Returns:
        speed of sound in [m/s]

    Examples:
        >> speed_of_sound(300.0)
    """
    gas_const = s_consts.R  # [J/mol*K]
    air_atomic_mass = air_atomic_molar_mass(["N2", "O2", "Ar", "CO2"])  # [g/mol]

    air_molecular_mass = (
        78 * air_atomic_mass["N2"]
        + 21 * air_atomic_mass["O2"]
        + 0.93 * air_atomic_mass["Ar"]
        + 0.07 * air_atomic_mass["CO2"]
    ) * 1e-5  # [kg/mol]
    spd_of_sound = np.sqrt(
        adiabatic_indx * temperature_K * gas_const / air_molecular_mass
    )
    return spd_of_sound  # [m/s]


def isentropic_relations(
    mach_1: float, adiabatic_indx: float = 1.4
) -> dict[str, float]:
    """
    Calculates isentropic relations

    Parameters:
        mach_1: pre-shock mach number

    Returns:
        dict: A dictionary containing:
            - pressure_s: pressure ratio (post-shock stagnation / pre-shock)
            - temperature_s: temperature ratio (post-shock stagnation / pre-shock)
            - density_s: density ratio (post-shock stagnation / pre-shock)
    Examples:
        >> isentropic_relations(3.0)

    Reference:
        Modern Compressible Flow With Historic Perspective, International
        Edition 4th (Anderson J., ISBN 978 1 260 57082 3)
    """
    gamma_minus = adiabatic_indx - 1
    gamma_ratio = gamma_minus / 2

    # Stagnation temperature (Eq. 3.28)
    temperature_s = 1 + gamma_ratio * mach_1**2

    # Stagnation pressure (Eq. 3.29)
    pressure_s = temperature_s ** (adiabatic_indx / gamma_minus)

    # Stagnation density (Eq. 3.30)
    density_s = temperature_s ** (1.0 / gamma_minus)

    isentropic_dict = {
        "pressure_s": pressure_s,
        "temperature_s": temperature_s,
        "density_s": density_s,
    }
    return isentropic_dict


def normal_shock_relations(
    mach_1: float, adiabatic_indx: float = 1.4
) -> dict[str, float]:
    """
    Calculates normal shock relations

    Parameters:
        mach_1: pre-shock mach number
        adiabatic_indx: adiabatic index, 1.4 (default)

    Returns:
        dict: A dictionary containing:
            - mach_2: post-shock mach number
            - pressure_r: pressure ratio (post-shock / pre-shock)
            - temperature_r: temperature ratio (post-shock / pre-shock)
            - density_r: density ratio (post-shock / pre-shock)
            - pressure_s: stagnation pressure ratio (post-shock / pre-shock)

    Reference:
        Normal Shock Wave - NASA (https://www.grc.nasa.gov/www/k-12/airplane/normal.html)
    """
    gamma_minus = adiabatic_indx - 1
    gamma_plus = adiabatic_indx + 1
    mach_11 = mach_1**2

    # Mach post-shock
    mach_2 = gamma_minus * mach_11 + 2
    mach_2 /= 2 * adiabatic_indx * mach_11 - gamma_minus
    mach_2 **= 0.5

    # Pressure ratio
    pressure_r = (2 * adiabatic_indx * mach_11 - gamma_minus) / gamma_plus

    # Temperature ratio
    temperature_r = 2 * adiabatic_indx * mach_11 - gamma_minus
    temperature_r *= gamma_minus * mach_11 + 2
    temperature_r /= gamma_plus**2 * mach_11

    # Density ratio
    density_r = gamma_plus * mach_11 / (gamma_minus * mach_11 + 2)

    # Stagnation pressure ratio
    pressure_s1 = gamma_plus / (2 * adiabatic_indx * mach_11 - gamma_minus)
    pressure_s1 **= 1 / gamma_minus
    pressure_s2 = gamma_plus * mach_11 / (gamma_minus * mach_11 + 2)
    pressure_s2 **= adiabatic_indx / gamma_minus
    pressure_s = pressure_s1 * pressure_s2

    normal_shock_dict = {
        "mach_2": mach_2,
        "pressure_r": pressure_r,
        "temperature_r": temperature_r,
        "density_r": density_r,
        "pressure_s": pressure_s,
    }
    return normal_shock_dict  # [ ]


def oblique_shock_relations(
    mach_1: float, shock_angle_deg: float, adiabatic_indx: float = 1.4
) -> dict[str, float]:
    """
    Calculates oblique shock relations for weak shocks

    Parameters:
        mach_1: pre-shock mach number
        shock_angle_deg: shock angle in degrees
        adiabatic_indx: adiabatic index, 1.4 (default)

    Returns:
        dict: A dictionary containing:
            - mach_2: post-shock mach number
            - pressure_r: pressure ratio (post-shock / pre-shock)
            - temperature_r: temperature ratio (post-shock / pre-shock)
            - density_r: density ratio (post-shock / pre-shock)
            - deflection_angle_degs: deflection angle in [degs]
            - mach_n1: normal pre-shock mach number
            - mach_n2: normal post-shock mach number
    Examples:
        >> oblique_shock_relations(3.0, 45.0)

    Reference:
        Modern Compressible Flow With Historic Perspective, International
        Edition 4th (Anderson J., ISBN 978 1 260 57082 3)
    """
    shock_angle = np.radians(shock_angle_deg)
    gamma_minus = adiabatic_indx - 1
    gamma_plus = adiabatic_indx + 1

    # Normal pre-shock mach number (Eq. 4.7)
    mach_n1 = mach_1 * np.sin(shock_angle)
    mach_n11 = mach_n1**2

    # Check if the normal mach is greater than 1
    if not mach_n1 > 1:
        return print("No shock, found")

    # Deflection angle (Eq. 4.17)
    tan_deflection_ang = 2 / np.tan(shock_angle)
    tan_deflection_ang *= mach_n11 - 1
    tan_deflection_ang /= mach_1**2 * (adiabatic_indx + np.cos(2 * shock_angle)) + 2
    deflection_angle_deg = np.degrees(np.arctan(tan_deflection_ang))

    # Density ratio (Eq. 4.8)
    density_r = gamma_plus * mach_n11
    density_r /= gamma_minus * mach_n11 + 2

    # Pressure ratio (Eq. 4.9)
    pressure_r = 2 * adiabatic_indx * (mach_n11 - 1) / gamma_plus + 1

    # Normal post-shock mach number (Eq. 4.10)
    mach_n22 = mach_n11 + 2 / gamma_minus
    mach_n22 /= 2 * adiabatic_indx * mach_n11 / gamma_minus - 1

    # Temperature ratio (Eq. 4.11)
    temperature_r = pressure_r * 1 / density_r

    # Post-shock mach number (Eq. 4.12)
    mach_2 = mach_n22**0.5
    mach_2 /= np.sin(np.radians(shock_angle_deg - deflection_angle_deg))

    oblique_shock_dict = {
        "mach_2": mach_2,
        "pressure_r": pressure_r,
        "temperature_r": temperature_r,
        "density_r": density_r,
        "deflection_angle_degs": deflection_angle_deg,
        "mach_n2": mach_n22**0.5,
        "mach_n1": mach_n1,
    }
    return oblique_shock_dict



"""
    Date:   03/26/2023
    Author: Martin E. Liza
    File:   optics.py
    Def:    Contains aero optics functions.
"""

from ambiance import Atmosphere
import numpy as np
import scipy.constants as s_consts
from haot import aerodynamics as aero
from haot import constants as constants_tables
from haot import quantum_mechanics as quantum
from haot import conversions


def gas_density(density_dict: dict[str, float]) -> dict[str, float]:
    """
    Calculates gas density in [particles/m^3]

    Parameters:
        species in [kg/m^3]

    Returns:
        species density in [particles/m^3]
    """
    gas_amu_weight = aero.air_atomic_molar_mass()  # [g/mol]
    gas_density = {
        i: (density_dict[i] * 10**3 * s_consts.N_A / gas_amu_weight[i])
        for i in density_dict
    }

    return gas_density  # [particles/m^3]


def index_of_refraction(gas_density_dict: dict[str, float]) -> dict[str, float]:
    """
    Calculates index of refraction

    Parameters:
        species in [kg/m^3]

    Returns:
        dict: A dictionary containing
            - dilute: dilute index of refraction
            - dense: dense index of refraction
    """
    pol_consts = constants_tables.polarizability()  # [m^3]
    density = gas_density(gas_density_dict)  # [particles/m3]
    n_const = {}  # [ ]

    for i in gas_density_dict:
        # Pol constants are in m^3
        alpha = conversions.polarizability_cgs_to_si(pol_consts[i] * 1e6)
        n_const[i] = alpha * density[i]  # (a_i N_i)

    # add all n_i
    temp = 0.0
    for i in n_const.values():
        temp += i

    n_return = {}
    n_return["dilute"] = 1 + temp / (2 * s_consts.epsilon_0)
    n_temp = temp / (3 * s_consts.epsilon_0)
    n_return["dense"] = ((2 * n_temp + 1) / (1 - n_temp)) ** 0.5

    return n_return


def dielectric_material_const(n_const):
    # n ~ sqrt(e_r)
    dielectric_const_0 = s_consts.epsilon_0  # [F/m]
    dielectric = {}
    dielectric["dilute"] = dielectric_const_0 * n_const["dilute"] ** 2
    dielectric["dense"] = dielectric_const_0 * n_const["dense"] ** 2
    return dielectric


def optical_path_length(n_solution, distance):
    OPL = {}
    OPL["dilute"] = n_solution["dilute"] * distance
    OPL["dense"] = n_solution["dense"] * distance
    # TODO: Missing implementation
    print("TODO: Missing this implementation")


def tropina_aproximation(vibrational_number, rotational_number, molecule):
    electron_mass = s_consts.m_e
    electron_charge = s_consts.e
    spectroscopy_const = constants_tables.spectroscopy_constants(molecule)
    # resonance_distance = omega_gi - omega
    # TODO: Missing implementation
    print("TODO: Missing this implementation")


def buldakov_expansion(
    vibrational_number: int, rotational_number: int, molecule: str
) -> float:
    """
    Calculates the Buldakov expansion

    Parameters:
        vibrational_number: vibrational quantum number (has to be positive)
        rotational_number: rotational quantum number (has to be positive)
        molecule: H2, N2, O2

    Returns:
        buldakov expansion in [m^3]

    Reference:
        Temperature Dependence of Polarizability of Diatomic Homonuclear
        Molecules (https://doi.org/10.1134/BF03355985)
    """
    # Load constants
    spectroscopy_const = constants_tables.spectroscopy_constants(molecule)
    derivative_const = constants_tables.buldakov_polarizability_derivatives_2016(
        molecule
    )
    be_we = spectroscopy_const["B_e"] / spectroscopy_const["omega_e"]

    # Dunham potential energy constants
    (a_0, a_1, a_2) = quantum.potential_dunham_coef_012(molecule)
    a_3 = quantum.potential_dunham_coeff_m(a_1, a_2, 3)

    rotational_degeneracy = rotational_number * (rotational_number + 1)
    vibrational_degeneracy = 2 * vibrational_number + 1

    # Split in terms
    tmp_1 = be_we
    tmp_1 *= -3 * a_1 * derivative_const["first"] + derivative_const["second"]
    tmp_1 *= vibrational_degeneracy
    tmp_1 *= 1 / 2

    tmp_2 = be_we**2
    tmp_2 *= derivative_const["first"]
    tmp_2 *= rotational_degeneracy
    tmp_2 *= 4

    tmp_31a = 7
    tmp_31a += 15 * vibrational_degeneracy**2
    tmp_31a *= a_1**3
    tmp_31a *= -3 / 8

    tmp_31b = 23
    tmp_31b += 39 * vibrational_degeneracy**2
    tmp_31b *= a_2
    tmp_31b *= a_1
    tmp_31b *= 1 / 4

    tmp_31c = 5
    tmp_31c += vibrational_degeneracy**2
    tmp_31c *= a_3
    tmp_31c *= -15 / 4

    tmp_31 = derivative_const["first"] * (tmp_31a + tmp_31b + tmp_31c)

    tmp_32a = 7
    tmp_32a += 15 * vibrational_degeneracy**2
    tmp_32a *= a_1**2
    tmp_32a *= 1 / 8

    tmp_32b = 5
    tmp_32b += vibrational_degeneracy**2
    tmp_32b *= a_2
    tmp_32b * --3 / 4

    tmp_32 = derivative_const["second"] * (tmp_32a + tmp_32b)

    tmp_33 = 7
    tmp_33 += 15 * vibrational_degeneracy**2
    tmp_33 *= a_1
    tmp_33 *= derivative_const["third"]
    tmp_33 *= -1 / 24

    tmp_3 = (tmp_31 + tmp_32 + tmp_33) * be_we**2

    tmp_41 = 1 - a_2
    tmp_41 *= 24
    tmp_41 += 27 * a_1 * (1 + a_1)
    tmp_41 *= derivative_const["first"]

    tmp_42 = 1 + 3 * a_1
    tmp_42 *= derivative_const["second"]
    tmp_42 *= -3

    tmp_43 = 1 / 8 * derivative_const["third"]

    tmp_4 = tmp_41 + tmp_42 + tmp_43
    tmp_4 *= rotational_degeneracy
    tmp_4 *= vibrational_degeneracy
    tmp_4 *= be_we**3

    return derivative_const["zeroth"] + tmp_1 + tmp_2 + tmp_3 + tmp_4


def kerl_polarizability_temperature(
    temperature_K: float, molecule: str, wavelength_nm: float
) -> float:
    """
    Calculates the polarizability using Kerl's extrapolation

    Parameters:
        temperature_K: reference temperature in [K]
        molecule: H2, N2, O2, Air
        wavelength_nm: signal's wavelength in [nm]

    Returns:
        polarizability in [m^3]

    Reference:
        Polarizability a(w,T,rho) of Small Molecules in the Gas Phase
        (https://doi.org/10.1002/bbpc.19920960517)

    Examples:
        >> kerl_polarizability_temperature(600.0, 'N2', 533.0)
    """

    # Check sizes
    mean_const = constants_tables.kerl_interpolation(molecule)
    angular_frequency = 2 * np.pi * s_consts.speed_of_light / (wavelength_nm * 1e-9)

    tmp = mean_const["c"] * temperature_K**2
    tmp += mean_const["b"] * temperature_K
    tmp += 1
    tmp *= mean_const["groundPolarizability"]
    tmp /= 1 - (angular_frequency / mean_const["groundFrequency"]) ** 2

    return tmp  # [m^3]


def atmospheric_index_of_refraction(
    altitude_m: float, vapor_pressure: float = 0.0
) -> float:
    """
    Calculates the atmospheric index of refraction as a function of altitude

    Parameters:
        altitude_m: altitude in [m]
        vapor_pressure: vapor pressure at given altitude in [mbar], 0.0 (default)
        temperature_K: reference temperature in [K]

    Returns:
        index of refraction in [ ]

    Reference:
        The constants in the equation for atmospheric refractive index at radio frequencies (https://ieeexplore.ieee.org/document/4051437)
    """
    atmospheric_prop = Atmosphere(altitude_m)
    temperature = atmospheric_prop.temperature  # [K]
    pressure = atmospheric_prop.pressure * 0.01  # [mbar]
    [K_1, K_2] = constants_tables.smith_atmospheric_constants()

    refractivity = K_2 * vapor_pressure / temperature
    refractivity += pressure
    refractivity *= K_1 / temperature
    refractivity *= 10**-6

    return refractivity + 1


def atmospheric_gladstone_dale(altitude=0.0, gas_composition_dict=None):
    gladstone_const = gladstone_dale()  # [m3/kg]

    if gas_composition_dict is None:
        gas_composition_dict = {}
        gas_composition_dict["N"] = 0.0
        gas_composition_dict["O"] = 0.0
        gas_composition_dict["NO"] = 0.0
        gas_composition_dict["N2"] = 0.79
        gas_composition_dict["O2"] = 0.21

    tmp = 0
    for i in gas_composition_dict.keys():
        tmp += gas_composition_dict[i] * gladstone_const[i]

    return tmp


def gladstone_dale(gas_density_dict=None):  # [kg/m3
    gas_amu_weight = aero.air_atomic_molar_mass()  # [g/mol]
    avogadro_number = s_consts.N_A  # [particles/mol]
    dielectric_const = s_consts.epsilon_0  # [F/m]
    pol_consts = constants_tables.polarizability()  # [m^3]

    # Convert CGS to SI
    pol_consts.update(
        {n: 4 * np.pi * dielectric_const * pol_consts[n] for n in pol_consts.keys()}
    )  # [Fm^2]

    # Calculate Gladstone dale
    gladstone_dale_const = {}
    for i in pol_consts:
        gladstone_dale_const[i] = (
            pol_consts[i]
            / (2 * dielectric_const)
            * (avogadro_number / gas_amu_weight[i])
            * 1e3
        )  # [m^3/kg]

    gladstone_dale_dict = {}
    if not gas_density_dict:
        return gladstone_dale_const  # [m^3/kg]
    else:
        gladstone_dale_dict["gladstone_dale"] = 0.0
        for i in gas_density_dict:
            gladstone_dale_dict[i] = (
                gladstone_dale_const[i] * gas_density_dict[i]
            ) / sum(gas_density_dict.values())
            gladstone_dale_dict["gladstone_dale"] += (
                gladstone_dale_const[i] * gas_density_dict[i]
            )
        gladstone_dale_dict["gladstone_dale"] /= sum(gas_density_dict.values())

        return gladstone_dale_dict  # [m^3/kg]
